@@ -45,6 +45,11 @@ struct _DummyDispatch
 	GDestroyNotify destroy;
 };
 
+gboolean jaw_debug = FALSE;
+
+GMutex *atk_bridge_mutex = NULL;
+GCond *atk_bridge_cond = NULL;
+
 GMutex *key_dispatch_mutex = NULL;
 GCond *key_dispatch_cond = NULL;
 static gint key_dispatch_result = KEY_DISPATCH_NOT_DISPATCHED;
@@ -86,9 +91,16 @@ jaw_idle_dispatch (GSource    *source,
 	return origin_g_idle_dispatch(source, callback, user_data);
 }
 
+static void jaw_exit_func ()
+{
+	_exit(0);
+}
+
 static gboolean
 jaw_load_atk_bridge (gpointer p)
 {
+	g_mutex_lock(atk_bridge_mutex);
+
 	if (!g_module_supported()) {
 		return NULL;
 	}
@@ -100,15 +112,20 @@ jaw_load_atk_bridge (gpointer p)
 	}
 
 	GVoidFunc dl_init;
-	GVoidFunc dl_shutdown;
-	if (!g_module_symbol( module, "gnome_accessibility_module_init", (gpointer*)&dl_init)
-			|| !g_module_symbol( module, "gnome_accessibility_module_shutdown", (gpointer*)&dl_shutdown)) {
+	if (!g_module_symbol( module, "gnome_accessibility_module_init", (gpointer*)&dl_init)) {
 		g_module_close(module);
 		return NULL;
 	}
 
 	(dl_init)();
-	g_atexit( dl_shutdown );
+	g_atexit( jaw_exit_func );
+
+	if (jaw_debug) {
+		printf("ATK Bridge has been loaded successfully\n");
+	}
+
+	g_cond_signal(atk_bridge_cond);
+	g_mutex_unlock(atk_bridge_mutex);
 
 	return FALSE;
 }
@@ -125,6 +142,11 @@ JNIEXPORT void JNICALL Java_org_GNOME_Accessibility_AtkWrapper_initNativeLibrary
 	// Hook up g_idle_dispatch
 	origin_g_idle_dispatch = g_idle_funcs.dispatch;
 	g_idle_funcs.dispatch = jaw_idle_dispatch;
+
+	const gchar* debug_env = g_getenv("JAW_DEBUG");
+	if (g_strcmp0(debug_env, "1") == 0) {
+		jaw_debug = TRUE;
+	}
 	
 	// Java app with GTK Look And Feel will load gail
 	// Set NO_GAIL to "1" to prevent gail from executing
@@ -142,6 +164,9 @@ JNIEXPORT void JNICALL Java_org_GNOME_Accessibility_AtkWrapper_initNativeLibrary
 
 	jaw_impl_init_mutex();
 
+	atk_bridge_mutex = g_mutex_new();
+	atk_bridge_cond = g_cond_new();
+
 	key_dispatch_mutex = g_mutex_new();
 	key_dispatch_cond = g_cond_new();
 
@@ -152,8 +177,16 @@ JNIEXPORT void JNICALL Java_org_GNOME_Accessibility_AtkWrapper_initNativeLibrary
 	gdk_threads_add_idle(jaw_dummy_idle_func, NULL);
 	g_idle_add(jaw_load_atk_bridge, NULL);
 
+	// We need to wait for the completion of the loading of ATK Bridge
+	// in order to ensure event listeners in ATK Bridge are properly
+	// registered before any emission of AWT event.
+	g_mutex_lock(atk_bridge_mutex);
+
 	GThread *main_loop_thread = g_thread_create( jni_main_loop,
 			(gpointer)main_loop, FALSE, NULL);
+
+	g_cond_wait(atk_bridge_cond, atk_bridge_mutex);
+	g_mutex_unlock(atk_bridge_mutex);
 }
 
 typedef enum _SigalType {
@@ -220,7 +253,7 @@ focus_notify_handler (gpointer p)
 	jobject global_ac = para->global_ac;
 
 	JNIEnv *jniEnv = jaw_util_get_jni_env();
-	JawImpl* jaw_impl = jaw_impl_find_instance(jniEnv, global_ac);
+	JawImpl* jaw_impl = jaw_impl_get_instance(jniEnv, global_ac);
 	
 	if (jaw_impl == NULL) {
 		free_callback_para(para);
@@ -502,7 +535,7 @@ window_state_change_handler (gpointer p)
 	jobject global_ac = para->global_ac;
 
 	JNIEnv *jniEnv = jaw_util_get_jni_env();
-	JawImpl* jaw_impl = jaw_impl_find_instance(jniEnv, global_ac);
+	JawImpl* jaw_impl = jaw_impl_get_instance(jniEnv, global_ac);
 	
 	if (jaw_impl == NULL) {
 		free_callback_para(para);
@@ -658,7 +691,7 @@ signal_emit_handler (gpointer p)
 		case Sig_Object_Active_Descendant_Changed:
 		{
 			jobject child_ac = (*jniEnv)->GetObjectArrayElement(jniEnv, args, 0);
-			JawImpl *child_impl = jaw_impl_find_instance(jniEnv, child_ac);
+			JawImpl *child_impl = jaw_impl_get_instance(jniEnv, child_ac);
 			if (!child_impl) {
 				break;
 			}
@@ -830,7 +863,7 @@ object_state_change_handler (gpointer p)
 	jobject global_ac = para->global_ac;
 
 	JNIEnv *jniEnv = jaw_util_get_jni_env();
-	JawImpl* jaw_impl = jaw_impl_find_instance(jniEnv, global_ac);
+	JawImpl* jaw_impl = jaw_impl_get_instance(jniEnv, global_ac);
 	
 	if (jaw_impl == NULL) {
 		free_callback_para(para);
@@ -871,6 +904,11 @@ component_added_handler (gpointer p)
 	JNIEnv *jniEnv = jaw_util_get_jni_env();
 	JawImpl* jaw_impl = jaw_impl_get_instance(jniEnv, global_ac);
 	
+	AtkObject* atk_obj = ATK_OBJECT(jaw_impl);
+	if (atk_object_get_role(atk_obj) == ATK_ROLE_TOOL_TIP) {
+		atk_object_notify_state_change(atk_obj, ATK_STATE_SHOWING, 1);
+	}
+
 	free_callback_para(para);
 
 	return FALSE;
@@ -884,6 +922,38 @@ JNIEXPORT void JNICALL Java_org_GNOME_Accessibility_AtkWrapper_componentAdded(
 	g_idle_add(component_added_handler, para);
 }
 
+static gboolean
+component_removed_handler (gpointer p)
+{
+	CallbackPara *para = (CallbackPara*)p;
+	jobject global_ac = para->global_ac;
+
+	JNIEnv *jniEnv = jaw_util_get_jni_env();
+	JawImpl* jaw_impl = jaw_impl_find_instance(jniEnv, global_ac);
+
+	if (!jaw_impl) {
+		free_callback_para(para);
+		return FALSE;
+	}
+
+	AtkObject* atk_obj = ATK_OBJECT(jaw_impl);
+	if (atk_object_get_role(atk_obj) == ATK_ROLE_TOOL_TIP) {
+		atk_object_notify_state_change(atk_obj, ATK_STATE_SHOWING, 0);
+	}
+
+	free_callback_para(para);
+
+	return FALSE;
+}
+
+JNIEXPORT void JNICALL Java_org_GNOME_Accessibility_AtkWrapper_componentRemoved(
+		JNIEnv *jniEnv, jclass jClass, jobject jAccContext) {
+	jobject global_ac = (*jniEnv)->NewGlobalRef(jniEnv, jAccContext);
+	CallbackPara *para = alloc_callback_para(global_ac);
+
+	g_idle_add(component_removed_handler, para);
+}
+
 static gboolean
 key_dispatch_handler (gpointer p)
 {
